#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <memory.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include "tine.h"
#include "gtweqm.h"

#ifdef WEB_GATEWAY
#define tagCurrent "Current = "
#define tagLifetime "Lifetime = "
#define tagEnergy "Energy = "
#define tagState "State = "
#define tagText "Text = "
#define tagBeamX "Beam X = "
#define tagBeamY "Beam Y = "
#define tagMonX "Monitor X = "
#define tagMonY "Monitor Y = "
#define tagTime "lung = "
#define tagShutter "shutter = "

#define month(s) \
  !strncmp(s,"Jan",3) ? 0 :\
  !strncmp(s,"Feb",3) ? 1 :\
  !strncmp(s,"Mar",3) ? 2 :\
  !strncmp(s,"Apr",3) ? 3 :\
  !strncmp(s,"May",3) ? 4 :\
  !strncmp(s,"Jun",3) ? 5 :\
  !strncmp(s,"Jul",3) ? 6 :\
  !strncmp(s,"Aug",3) ? 7 :\
  !strncmp(s,"Sep",3) ? 8 :\
  !strncmp(s,"Oct",3) ? 9 :\
  !strncmp(s,"Nov",3) ? 10 :\
  !strncmp(s,"Dec",3) ? 11 : 0

typedef struct 
{
  char *tag;
  float *fval;
  int *ival;
  char *sval;
  int siz;
  time_t *ts;
} DorisDataRecord;
DorisDataRecord dorisRecords[] = 
{
  {tagCurrent, &g_current, NULL, NULL, 1, &ts_current},
  {tagLifetime, &g_lifetime, NULL, NULL, 1, &ts_lifetime},
  {tagEnergy, &g_energy, NULL, NULL, 1, &ts_energy},
  {tagState, NULL, NULL, g_state.name, 1, &ts_state},
  {tagText, NULL, NULL, g_text.name, 1, &ts_text},
  {tagBeamX, g_bpmx, NULL, NULL, 38, &ts_bpmx},
  {tagBeamY, g_bpmy, NULL, NULL, 38, &ts_bpmy},
  {tagMonX, g_monx, NULL, NULL, 38, &ts_monx},
  {tagMonY, g_mony, NULL, NULL, 38, &ts_mony},
};
DorisDataRecord timeRecord = 
{
  tagTime, NULL, &timeRemaining, NULL, 1, &ts_time
};
DorisDataRecord shutterRecord = 
{
  tagShutter, NULL, g_beamShutter, NULL, 34, &ts_shutter
};
int nDorisRecords = sizeof(dorisRecords)/sizeof(DorisDataRecord);
int parseOutput(const char *tag,char *buf,int siz,float *fval,int *ival,char *sval,time_t *ts)
{
  int i, p=24;
  char sbuf[128], *c,*end=NULL;
  time_t t = 0;
  struct tm tms;

  if (buf == NULL) return argument_list_error;
  if (fval != NULL)
  {
    if ((c=strstr(buf,tag)) != NULL)
    {
      c += strlen(tag);
      for (i=0; i<siz; i++)
      {
        fval[i] = (float)strtod(c,&c);
        c++;
      }
    }
    goto out;
  }
  if (ival != NULL)
  {
    if ((c=strstr(buf,tag)) != NULL)
    {
      c += strlen(tag);
      for (i=0; i<siz; i++)
      {
        ival[i] = strtol(c,&c,10);
        c++;
      }
    }
    goto out;
  }
  if (sval != NULL)
  {
    strcpy(sval,"-1"); // let the incoming string overwrite this :
    if ((c=strstr(buf,tag)) != NULL)
    {
      c += strlen(tag);
      strncpy(sbuf,c,127); sbuf[127] = 0;
      if ((end=strstr(sbuf," at")) != NULL) *end = 0;
      strncpy(sval,sbuf,64);
    }
    goto out;
  }
out:
  if (c == NULL) return string_expected; // nothing found !
  if ((end=strstr(c," at")) == NULL) return string_expected; // corrupted string !
  if (ts != NULL)
  {
    end += 4;
    while (end[p] != '2' && end[p] != '<') p++; // dst string ?
    tms.tm_year = atoi(&end[p]) - 1900;
    tms.tm_mon = month(&end[4]);
    tms.tm_mday = atoi(&end[8]);
    tms.tm_hour = atoi(&end[11]);
    tms.tm_min = atoi(&end[14]);
    tms.tm_sec = atoi(&end[17]);
    tms.tm_isdst = -1;
    *ts = mktime(&tms);
  }
  return 0;
}
#define contentLengthTag "Content-Length: "
int getContentSize(char *buf)
{
  char *c = strstr(buf,contentLengthTag);
  if (c == NULL) return 0;
  c += strlen(contentLengthTag);
  return atoi(c);
}
int parseBeamShutter(char *buf,int *shttr,time_t *ts)
{
  int cc = net_read_error;

  if (buf == NULL) return argument_list_error;
  if (shttr != NULL)
  {
    cc = parseOutput(shutterRecord.tag,
                     buf,
                     34,
                     NULL,
                     shttr,
                     NULL,
                     ts);
  }
  return cc;
}
int parseTimeRemaining(char *buf,int *t,time_t *ts)
{
  int i, p=24;
  char *c=NULL,*end=NULL;
  struct tm tms;

  if (buf == NULL) return argument_list_error;
  if (t != NULL)
  {
    if ((c=strstr(buf,timeRecord.tag)) != NULL)
    {
      c += strlen(timeRecord.tag);
      *t = strtol(c,&c,10);
      c++;
    }
    goto out;
  }
out:
  if (c == NULL) return string_expected; // nothing found !
  if ((end=strstr(c," at")) == NULL) return string_expected; // corrupted string !
  if (ts != NULL)
  {
    end += 4;
    while (end[p] != '2' && end[p] != '<') p++; // dst string ?
    tms.tm_year = atoi(&end[p]) - 1900;
    tms.tm_mon = month(&end[4]);
    tms.tm_mday = atoi(&end[8]);
    tms.tm_hour = atoi(&end[11]);
    tms.tm_min = atoi(&end[14]);
    tms.tm_sec = atoi(&end[17]);
    tms.tm_isdst = -1;
    *ts = mktime(&tms);
  }
  return 0;
}

int getDorisTimeRemaining(void)
{
  static char url[256];
  static char hostname[256];
  static char filename[200];
  char buf[4096];
  static int  port = 8080;
  int sck = -1, i, n, p, mode = 1, cc = 0;
  static struct hostent *host;
  static struct sockaddr_in  serverAddr;
  static int initDone = 0;
  fd_set rset;
  struct timeval to;
  
  if (!initDone)
  {
    // fill in the destinations :
    filename[0] = 0;
    strcat(filename,"http://webmex.desy.de:8080/servlet/WebRepeater?linkname=%2FDORIS%2FDORISDATA%2FTimeRemaining");
    sprintf(url,"GET %s HTTP/1.0\n\n",filename);
    sprintf(hostname,"mexgate.desy.de");
    // fill in the server sockaddr:
    memset(&serverAddr,0,sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port = htons(port);
    if ((host=gethostbyname(hostname)) == NULL &&
        (host=gethostbyaddr(hostname,strlen(hostname),AF_INET)) == NULL)
    {
      {
        perror("gethostbyname");
        feclog("cannot find host %s",hostname);
        exit(1);
      }
    }
    memcpy(&serverAddr.sin_addr,host->h_addr,sizeof(struct in_addr));
    initDone = -1;
  }

  if ((sck=socket(AF_INET,SOCK_STREAM,0)) < 0 )
  {
    perror("socket");
    feclog("cannot get tcp socket !");
    return tcp_socket_error;
  }
  if (connect(sck,(struct sockaddr *)&serverAddr,sizeof(serverAddr)) < 0)
  {
    perror("connect");
    feclog("cannot connect to host %s",hostname);
    ccerr(tcp_connect_error);
  }

  n = write(sck,url,strlen(url));
  for (i =0, n = 0; i<3; i++)
  {
    n += read(sck,&buf[n],4096-n);
    p = getContentSize(buf);
    if (n >= p) break;
  }
  if (n > 0) 
  {
    if (n > 4095) n = 4095;
    buf[n] = 0;
    cc = parseTimeRemaining(buf,timeRecord.ival,timeRecord.ts);
  }
  else
  {
    cc = net_read_error;
  }

err:
  if (sck > 0) close(sck);

  return cc;
}

int getMainBeamShutter(void)
{
  static char url[256];
  static char hostname[256];
  static char filename[200];
  char buf[4096];
  static int  port = 8080;
  int sck = -1, i, n, p, mode = 1, cc = 0;
  static struct hostent *host;
  static struct sockaddr_in  serverAddr;
  static int initDone = 0;
  fd_set rset;
  struct timeval to;
  
  if (!initDone)
  {
    // fill in the destinations :
    filename[0] = 0;
    strcat(filename,"http://webmex.desy.de:8080/servlet/WebRepeater?linkname=%2FDORIS%2FHASYBEAMLINES%2FW2%2FMAIN_BEAMSHUTTER");
    sprintf(url,"GET %s HTTP/1.0\n\n",filename);
    sprintf(hostname,"mexgate.desy.de");
    // fill in the server sockaddr:
    memset(&serverAddr,0,sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port = htons(port);
    if ((host=gethostbyname(hostname)) == NULL &&
        (host=gethostbyaddr(hostname,strlen(hostname),AF_INET)) == NULL)
    {
      {
        perror("gethostbyname");
        feclog("cannot find host %s",hostname);
        exit(1);
      }
    }
    memcpy(&serverAddr.sin_addr,host->h_addr,sizeof(struct in_addr));
    initDone = -1;
  }

  if ((sck=socket(AF_INET,SOCK_STREAM,0)) < 0 )
  {
    perror("socket");
    feclog("cannot get tcp socket !");
    return tcp_socket_error;
  }
  if (connect(sck,(struct sockaddr *)&serverAddr,sizeof(serverAddr)) < 0)
  {
    perror("connect");
    feclog("cannot connect to host %s",hostname);
    ccerr(tcp_connect_error);
  }

  n = write(sck,url,strlen(url));
  for (i =0, n = 0; i<3; i++)
  {
    n += read(sck,&buf[n],4096-n);
    p = getContentSize(buf);
    if (n >= p) break;
  }
  if (n > 0) 
  {
    if (n > 4095) n = 4095;
    buf[n] = 0;
    cc = parseBeamShutter(buf,shutterRecord.ival,shutterRecord.ts);
  }
  else
  {
    cc = net_read_error;
  }

err:
  if (sck > 0) close(sck);

  return cc;
}
#endif
LINK links[] = 
{
  "/DORIS/ARCHIVER/keyword","CurDC","",CF_FLOAT,1,-1,0,(BYTE *)&g_current,&ts_current,
  "/DORIS/ARCHIVER/keyword","TauDC","",CF_FLOAT,1,-1,0,(BYTE *)&g_lifetime,&ts_lifetime,
  "/DORIS/ARCHIVER/keyword","Energy","",CF_FLOAT,1,-1,0,(BYTE *)&g_energy,&ts_energy,
  "/DORIS/ARCHIVER/keyword","DOSTATE","",CF_NAME32,1,-1,0,(BYTE *)g_state.name,&ts_state,
  "/DORIS/ARCHIVER/keyword","DoTVText","",CF_TEXT,64,-1,0,(BYTE *)g_text.name,&ts_text,
  "/DORIS/ARCHIVER/W2.1","BLBeamPos.X","",CF_FLOAT,38,-1,0,(BYTE *)g_bpmx,&ts_bpmx,
  "/DORIS/ARCHIVER/W2.1","BLBeamPos.Y","",CF_FLOAT,38,-1,0,(BYTE *)g_bpmy,&ts_bpmy,
  "/DORIS/ARCHIVER/W2.1","BLMonPos.X","",CF_FLOAT,38,-1,0,(BYTE *)g_monx,&ts_monx,
  "/DORIS/ARCHIVER/W2.1","BLMonPos.Y","",CF_FLOAT,38,-1,0,(BYTE *)g_mony,&ts_mony,
  "/DORIS/ARCHIVER/keyword","TimeRemaining","",CF_INT32,1,-1,0,(BYTE *)&timeRemaining,&ts_time,
  "/DORIS/ARCHIVER/W2","MAIN_BEAMSHUTTER","",CF_INT32,34,-1,0,(BYTE *)g_beamShutter,&ts_shutter,
};
int nlinks = sizeof(links)/sizeof(links[0]);

void cbLinks(int id,int cc)
{
  LINK *lnk;

  if (cc != 0)
  {
    return;
  }
  if (id < 0 || id >= nlinks)
  {
    return;
  }
  lnk = &links[id];
  *lnk->timestamp = GetCurrentDataTimeStamp(lnk->id);
}
int startLinks(void)
{
  int i,fsiz,cc = 0;
  DTYPE dout;
 
  for (i=0; i<nlinks; i++)
  {
    dout.dArrayLength = links[i].siz;
    dout.dFormat = links[i].fmt;
    strncpy(dout.dTag,links[i].tag,TAG_NAME_SIZE);
    fsiz = fmtsizeof(links[i].fmt);
    dout.data.bptr = links[i].data;
    if ((links[i].dout=(DTYPE *)calloc(1,sizeof(DTYPE))) == NULL)
    {
      feclog("cannot allocate link memory !");
      printf("cannot allocate link memory !\n");
      exit(1);
    }
    memcpy(links[i].dout,&dout,sizeof(DTYPE));
    feclog("starting link to %s %s",links[i].dev,links[i].prp);
    links[i].id = AttachLinkEx(links[i].dev,links[i].prp,links[i].dout,NULL,CA_READ,1000,cbLinks,CM_TIMER,i);
    if (links[i].id < 0) 
    {
      cc = -links[i].id;
      dbglog("error starting link to %s %s : %s",
        links[i].dev,links[i].prp,erlst[cc]);
    }
  }
  return cc;
}
int getDorisParameters(void)
{
  int sck = -1, i, n, p, mode = 1, cc = 0;
  static int initDone = 0;
  
  if (!initDone)
  {
    initDone = -1;
  }

# ifdef WEB_GATEWAY
    for (i=0; i<nDorisRecords; i++)
      cc = parseOutput(dorisRecords[i].tag,
                     buf,
                     dorisRecords[i].siz,
                     dorisRecords[i].fval,
                     dorisRecords[i].ival,
                     dorisRecords[i].sval,
                     dorisRecords[i].ts);
# endif
err:

  return cc;
}
